from collections import namedtuple

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import ProbesQuery, probes_table
from diamond_miner.typing import IPNetwork


class GetProbes(ProbesQuery):
    """
    Return the cumulative number of probes sent at a specified round.

    >>> from diamond_miner.test import addr_to_string, url
    >>> row = GetProbes.Row(*GetProbes(round_eq=1).execute(url, 'test_nsdi_example')[0])
    >>> row.protocol
    1
    >>> addr_to_string(row.dst_prefix)
    '200.0.0.0'
    >>> sorted(row.probes_per_ttl)
    [(1, 6), (2, 6), (3, 6), (4, 6)]
    >>> row = GetProbes.Row(*GetProbes(round_eq=2).execute(url, 'test_nsdi_example')[0])
    >>> sorted(row.probes_per_ttl)
    [(1, 11), (2, 18), (3, 18), (4, 18)]
    >>> row = GetProbes.Row(*GetProbes(round_eq=3).execute(url, 'test_nsdi_example')[0])
    >>> sorted(row.probes_per_ttl)
    [(1, 11), (2, 20), (3, 27), (4, 27)]
    >>> GetProbes(round_eq=4).execute(url, 'test_nsdi_example')
    []
    """

    Row = namedtuple("Row", "protocol,dst_prefix,probes_per_ttl")

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_eq, "`round_eq` must be specified."
        return f"""
        SELECT
            probe_protocol,
            probe_dst_prefix,
            groupArray((probe_ttl, cumulative_probes)) AS probes_per_ttl
        FROM {probes_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (probe_protocol, probe_dst_prefix)
        """


class GetProbesDiff(ProbesQuery):
    """
    Return the number of probes sent at a specific round and at the previous round.

    >>> from diamond_miner.test import addr_to_string, url
    >>> row = GetProbesDiff.Row(*GetProbesDiff(round_eq=1).execute(url, 'test_nsdi_example')[0])
    >>> row.protocol
    1
    >>> addr_to_string(row.dst_prefix)
    '200.0.0.0'
    >>> sorted(row.probes_per_ttl)
    [(1, 6, 0), (2, 6, 0), (3, 6, 0), (4, 6, 0)]
    >>> row = GetProbesDiff.Row(*GetProbesDiff(round_eq=2).execute(url, 'test_nsdi_example')[0])
    >>> sorted(row.probes_per_ttl)
    [(1, 11, 6), (2, 18, 6), (3, 18, 6), (4, 18, 6)]
    >>> row = GetProbesDiff.Row(*GetProbesDiff(round_eq=3).execute(url, 'test_nsdi_example')[0])
    >>> sorted(row.probes_per_ttl)
    [(1, 11, 11), (2, 20, 18), (3, 27, 18), (4, 27, 18)]
    >>> GetProbesDiff(round_eq=4).execute(url, 'test_nsdi_example')
    []
    """

    Row = namedtuple("Row", "protocol,dst_prefix,probes_per_ttl")

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert self.round_eq
        return f"""
        SELECT
            current.probe_protocol,
            current.probe_dst_prefix,
            groupArray((current.probe_ttl, current.cumulative_probes, previous.cumulative_probes)) AS probes_per_ttl
        FROM {probes_table(measurement_id)} AS current
        LEFT JOIN (
            SELECT
                probe_protocol,
                probe_dst_prefix,
                probe_ttl,
                cumulative_probes
            FROM {probes_table(measurement_id)}
            WHERE round = {self.round_eq - 1}
            -- TODO: Filter on subset for greater performance?
        ) AS previous
        ON current.probe_protocol = previous.probe_protocol
        AND current.probe_dst_prefix = previous.probe_dst_prefix
        AND current.probe_ttl = previous.probe_ttl
        WHERE {self.filters(subset)}
        GROUP BY (current.probe_protocol, current.probe_dst_prefix)
        """



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.fragments import date_time
from diamond_miner.queries.query import (
    ResultsQuery,
    StoragePolicy,
    flows_table,
    results_table,
)
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreateFlowsView(ResultsQuery):
    """Create the flows view."""

    PRIMARY_KEY = "probe_protocol, probe_src_addr, probe_dst_prefix"
    SORTING_KEY = "probe_protocol, probe_src_addr, probe_dst_prefix, probe_dst_addr, probe_src_port, probe_dst_port"

    storage_policy: StoragePolicy = StoragePolicy()

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert subset == UNIVERSE_SUBSET, "subset not allowed for this query"
        return f"""
        CREATE MATERIALIZED VIEW IF NOT EXISTS {flows_table(measurement_id)}
        ENGINE = AggregatingMergeTree
        ORDER BY ({self.SORTING_KEY})
        PRIMARY KEY ({self.PRIMARY_KEY})
        TTL {date_time(self.storage_policy.archive_on)} TO VOLUME '{self.storage_policy.archive_to}'
        SETTINGS storage_policy = '{self.storage_policy.name}'
        AS SELECT
            round,
            {self.SORTING_KEY},
            -- The duplication of the round column here allows us to drop `round`
            -- from the GROUP BY clause when querying the view, and still get the
            -- round information in the replies. This is useful for computing
            -- inter-round links.
            groupUniqArrayState((round, probe_ttl, reply_src_addr)) AS replies
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (round, {self.SORTING_KEY})
        """



from dataclasses import fields
from typing import Any, Dict


def common_parameters(from_dataclass: Any, to_dataclass: Any) -> Dict[str, Any]:
    to_params = {field.name for field in fields(to_dataclass)}
    return {
        field.name: getattr(from_dataclass, field.name)
        for field in fields(from_dataclass)
        if field.name in to_params
    }



from typing import Iterable, Iterator

from diamond_miner.defaults import (
    DEFAULT_PROBE_DST_PORT,
    DEFAULT_PROBE_SRC_PORT,
    PROTOCOLS,
    UNIVERSE_SUBSET,
)
from diamond_miner.logging import logger
from diamond_miner.queries import GetProbesDiff, InsertMDAProbes
from diamond_miner.queries.delete_probes import DeleteProbes
from diamond_miner.typing import FlowMapper, IPNetwork, Probe


def mda_probes(
    url: str,
    measurement_id: str,
    previous_round: int,
    mapper_v4: FlowMapper,
    mapper_v6: FlowMapper,
    probe_src_port: int = DEFAULT_PROBE_SRC_PORT,
    probe_dst_port: int = DEFAULT_PROBE_DST_PORT,
    adaptive_eps: bool = False,
    target_epsilon: float = 0.05,
    subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
) -> Iterator[Probe]:
    """
    Compute the probes to send given the previously discovered links.
    """
    next_round = previous_round + 1

    # 1) Delete eventual pre-existing probes.
    DeleteProbes(round_eq=next_round).execute(url, measurement_id, subsets)

    # 2) Compute and save the new probes.
    InsertMDAProbes(
        adaptive_eps=adaptive_eps,
        round_leq=previous_round,
        filter_partial=True,
        filter_virtual=True,
        filter_inter_round=True,
        target_epsilon=target_epsilon,
    ).execute(url, measurement_id, subsets)

    # 3) Generate the probe tuples for the next round.
    rows = GetProbesDiff(round_eq=next_round).execute_iter(url, measurement_id, subsets)
    for row in rows:
        row = GetProbesDiff.Row(*row)
        for probe in row_to_probes(
            row,
            mapper_v4,
            mapper_v6,
            probe_src_port,
            probe_dst_port,
        ):
            # TEMP: Log prefixes that overflows the port number.
            if probe[1] > (2 ** 16 - 1):
                logger.warning("Port overflow for %s", row)
                break
            yield probe


def row_to_probes(
    row: GetProbesDiff.Row,
    mapper_v4: FlowMapper,
    mapper_v6: FlowMapper,
    probe_src_port: int,
    probe_dst_port: int,
) -> Iterator[Probe]:
    dst_prefix_int = int(row.dst_prefix)
    mapper = mapper_v4 if row.dst_prefix.ipv4_mapped else mapper_v6
    protocol_str = PROTOCOLS[row.protocol]

    for ttl, total_probes, already_sent in row.probes_per_ttl:
        for flow_id in range(already_sent, total_probes):
            addr_offset, port_offset = mapper.offset(
                flow_id=flow_id, prefix=dst_prefix_int
            )
            yield (
                dst_prefix_int + addr_offset,
                probe_src_port + port_offset,
                probe_dst_port,
                ttl,
                protocol_str,
            )



from ipaddress import IPv4Network, IPv6Network, ip_network
from typing import Iterable, Iterator, List, Optional, Sequence, Tuple, Union

from diamond_miner.defaults import (
    DEFAULT_PREFIX_LEN_V4,
    DEFAULT_PREFIX_LEN_V6,
    DEFAULT_PREFIX_SIZE_V4,
    DEFAULT_PREFIX_SIZE_V6,
    DEFAULT_PROBE_DST_PORT,
    DEFAULT_PROBE_SRC_PORT,
)
from diamond_miner.grid import ParameterGrid
from diamond_miner.mappers import SequentialFlowMapper
from diamond_miner.typing import FlowMapper, Probe


def count_prefixes(
    prefixes: Iterable[str],
    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4,
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6,
) -> int:
    """
    Count the number of prefixes yielded by a list of prefixes.

    :param prefixes: A list of IPv4/v6 prefixes.
    :param prefix_len_v4: The target prefix length for v4 prefixes.
    :param prefix_len_v6: The target prefix length for v6 prefixes.
    :return: The number of prefixes yielded by the prefix list.

    >>> count_prefixes(["8.8.4.0/24", "8.8.8.0/24"])
    2
    >>> count_prefixes(["0.0.0.0/22"])
    4
    >>> count_prefixes(["0.0.0.0/0"])
    16777216
    >>> count_prefixes(["2001::/48"])
    65536
    >>> count_prefixes(["0.0.0.0/32"], prefix_len_v4=24)
    Traceback (most recent call last):
        ...
    ValueError: prefix length must be <= 24
    """
    count = 0
    for prefix in prefixes:
        network = ip_network(prefix)
        if isinstance(network, IPv4Network):
            prefix_len = prefix_len_v4
        else:
            prefix_len = prefix_len_v6
        if network.prefixlen > prefix_len:
            raise ValueError(f"prefix length must be <= {prefix_len}")
        count += 2 ** (prefix_len - network.prefixlen)
    return count


def split_prefix(
    prefix: str, prefix_len_v4: int, prefix_len_v6: int
) -> Iterator[Tuple[int, int, int]]:
    network = ip_network(prefix.strip())
    if isinstance(network, IPv4Network):
        # We add 0xFFFF00000000 to convert the network address
        # to an IPv4-mapped IPv6 address.
        prefix_size = 2 ** (32 - prefix_len_v4)
        for x in subnets(network, prefix_len_v4):
            yield 4, x + 0xFFFF00000000, prefix_size
    else:
        prefix_size = 2 ** (128 - prefix_len_v6)
        for x in subnets(network, prefix_len_v6):
            yield 6, x, prefix_size


def subnets(network: Union[IPv4Network, IPv6Network], new_prefix: int) -> Sequence[int]:
    """
    Faster version of :py:meth:`ipaddress.IPv4Network.subnets`.
    Returns only the network address as an integer.

    >>> from ipaddress import ip_network
    >>> list(subnets(ip_network("0.0.0.0/0"), new_prefix=2))
    [0, 1073741824, 2147483648, 3221225472]
    >>> subnets(ip_network("0.0.0.0/32"), new_prefix=24)
    Traceback (most recent call last):
        ...
    ValueError: new prefix must be longer
    """
    if new_prefix < network.prefixlen:
        raise ValueError("new prefix must be longer")
    start = int(network.network_address)
    end = int(network.broadcast_address) + 1
    step = (int(network.hostmask) + 1) >> (new_prefix - network.prefixlen)
    return range(start, end, step)


def probe_generator(
    prefixes: Sequence[Tuple[str, str]],  # /32 or / 128 if nothing specified
    flow_ids: Sequence[int] = range(6),
    ttls: Sequence[int] = range(1, 33),
    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4,
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6,
    probe_src_port: int = DEFAULT_PROBE_SRC_PORT,
    probe_dst_port: int = DEFAULT_PROBE_DST_PORT,
    mapper_v4: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V4),
    mapper_v6: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V6),
    seed: Optional[int] = None,
) -> Iterator[Probe]:
    """
    Generate a probe for each prefix, flow ID and TTL, in a random order.

    :param prefixes: A list of (prefix, protocol) tuples. The protocol can be ``icmp``, ``icmp6`` or ``udp``.
    :param flow_ids: The flow IDs to probe.
    :param ttls: The TTLs to probe.
    :param prefix_len_v4: The prefix length to which the IPv4 prefixes will be split to.
    :param prefix_len_v6: The prefix length to which the IPv6 prefixes will be split to.
    :param probe_src_port: The minimum source port of the probes (can be incremented by the flow mapper).
    :param probe_dst_port: The destination port of the probes (constant).
    :param mapper_v4: The flow mapper for IPv4 probes.
    :param mapper_v6: The flow mapper for IPv6 probes.
    :param seed: The seed of the random permutation (two calls with the same seed will yield the probes in the same order).

    This function is very versatile, it can generate Tokyo-Ping :cite:`pelsser2013paris`,
    Paris-Traceroute :cite:`augustin2006avoiding` or Yarrp-like :cite:`beverly2016yarrp` probes.

    For ICMP probes, the source port is encoded by caracal in the checksum field of the ICMP header
    which is generally used by routers for per-flow load-balancing.

    .. code-block:: python

        # Example: ICMP ping towards Google DNS servers with 2 flows per address and a TTL of 32.
        prefixes = [("8.8.8.8/32", "icmp"), ("2001:4860:4860::8888/128", "icmp6")]
        prefix_len_v4 = 32
        prefix_len_v6 = 128
        flow_ids = range(2)
        ttls = [32]

        # When given a prefix size of 1, the sequential flow mapper will only vary the port.
        mapper_v4 = SequentialFlowMapper(prefix_size=1)
        mapper_v6 = SequentialFlowMapper(prefix_size=1)

        # Example: ICMP ping towards 1.0.0.0/24 with 1 flow per address and a TTL of 32.
        prefixes = [("1.0.0.0/24", "icmp")]
        prefix_len_v4 = 32 # The generator will cut the /24 in 256 /32.
        flow_ids = range(1)
        ttls = [32]
        # Same flow mappers as above.

        # Example: UDP traceroute towards 1.0.0.0/24 with 2 flows per address.
        # 256 addresses * 2 flows * 30 TTLs = 15,360 probes.
        prefixes = [("1.0.0.0/24", "udp")]
        prefix_len_v4 = 32 # The generator will cut the /24 in 256 /32.
        flow_ids = range(2)
        ttls = range(2, 32)
        # Same flow mappers as above.

        # Example: UDP traceroute towards 1.0.0.0/24 with 6 flows **per prefix**.
        # 1 prefix * 6 flows * 30 TTLs = 180 probes.
        prefixes = [("1.0.0.0/24", "udp")]
        prefix_len_v4 = 24 # We want to target the prefix, not its individual addresses.
        flow_ids = range(6)
        ttls = range(2, 32)

        # The random flow mapper will assign a random destination (in the /24) to each flow.
        mapper_v4 = RandomFlowMapper(prefix_size=256)
    """
    prefixes_: List[Tuple[int, int, int, str]] = []
    for prefix, protocol in prefixes:
        for af, subprefix, subprefix_size in split_prefix(
            prefix, prefix_len_v4, prefix_len_v6
        ):
            prefixes_.append((af, subprefix, subprefix_size, protocol))

    grid = ParameterGrid(prefixes_, ttls, flow_ids).shuffled(seed=seed)

    for (af, subprefix, subprefix_size, protocol), ttl, flow_id in grid:
        mapper = mapper_v4 if af == 4 else mapper_v6
        addr_offset, port_offset = mapper.offset(flow_id=flow_id, prefix=subprefix)
        yield subprefix + addr_offset, probe_src_port + port_offset, probe_dst_port, ttl, protocol


def probe_generator_by_flow(
    prefixes: Iterable[
        Tuple[str, str, Iterable[int]]
    ],  # /32 or / 128 if nothing specified
    flow_ids: Sequence[int] = range(6),
    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4,
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6,
    probe_src_port: int = DEFAULT_PROBE_SRC_PORT,
    probe_dst_port: int = DEFAULT_PROBE_DST_PORT,
    mapper_v4: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V4),
    mapper_v6: FlowMapper = SequentialFlowMapper(DEFAULT_PREFIX_SIZE_V6),
    seed: Optional[int] = None,
) -> Iterator[Probe]:
    """
    Generate a probe for each prefix, flow id and TTL, in a random order.
    This function differs from :func:`probe_generator` in two aspects:

        * The TTLs are specified for each prefixes, and not globally.
        * All the probes for a given prefix and flow id are generated sequentially.

    The parameters and output are identical to :func:`probe_generator`,
    excepted for ``prefixes`` which is a list of (prefix, protocol, TTLs) tuples,
    and the absence of the ``ttls`` parameter.
    """
    prefixes_: List[Tuple[int, int, int, str, Iterable[int]]] = []
    for prefix, protocol, ttls in prefixes:
        for af, subprefix, subprefix_size in split_prefix(
            prefix, prefix_len_v4, prefix_len_v6
        ):
            prefixes_.append((af, subprefix, subprefix_size, protocol, ttls))

    grid = ParameterGrid(prefixes_, flow_ids).shuffled(seed=seed)

    for (af, subprefix, subprefix_size, protocol, ttls), flow_id in grid:
        mapper = mapper_v4 if af == 4 else mapper_v6
        for ttl in ttls:
            addr_offset, port_offset = mapper.offset(flow_id=flow_id, prefix=subprefix)
            yield subprefix + addr_offset, probe_src_port + port_offset, probe_dst_port, ttl, protocol



import asyncio
import json
import os
from asyncio import Semaphore
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import reduce
from typing import AsyncIterator, Iterable, Iterator, List, Optional, Sequence, Tuple

import httpx
from aioch import Client as AsyncClient
from clickhouse_driver import Client
from clickhouse_driver.errors import ServerException

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.logging import logger
from diamond_miner.queries.fragments import (
    and_,
    eq,
    geq,
    ip_eq,
    ip_in,
    leq,
    lt,
    not_,
    or_,
)
from diamond_miner.timer import LoggingTimer
from diamond_miner.typing import IPNetwork

CH_QUERY_SETTINGS = {
    "max_block_size": 128_000,
    # Avoid timeout in case of a slow connection
    "connect_timeout": 1000,
    "send_timeout": 6000,
    "receive_timeout": 6000,
    # https://github.com/ClickHouse/ClickHouse/issues/18406
    "read_backoff_min_latency_ms": 100_000,
}


@asynccontextmanager
async def async_client(url: str) -> AsyncClient:
    c = AsyncClient.from_url(url)
    try:
        yield c
    finally:
        await c.disconnect()


@contextmanager
def client(url: str) -> Client:
    c = Client.from_url(url)
    try:
        yield c
    finally:
        c.disconnect()


def flows_table(measurement_id: str) -> str:
    return f"flows__{measurement_id}".replace("-", "_")


def links_table(measurement_id: str) -> str:
    return f"links__{measurement_id}".replace("-", "_")


def prefixes_table(measurement_id: str) -> str:
    return f"prefixes__{measurement_id}".replace("-", "_")


def probes_table(measurement_id: str) -> str:
    return f"probes__{measurement_id}".replace("-", "_")


def results_table(measurement_id: str) -> str:
    return f"results__{measurement_id}".replace("-", "_")


class AddrType(str, Enum):
    """ClickHouse IPv6 type."""

    IPv6 = "IPv6"
    """
    IPv6 stored as FixedString(16) internally.
    Slow with clickhouse-driver as it will call :func:`ipaddress.IPv6Address` on each address.
    """

    String = "String"
    """
    String representation of the internal FixedString.
    With clickhouse-driver each address will be decoded as Python string.
    """

    FixedString = "FixedString"
    """
    Native FixedString(16) representation.
    Fastest with clickhouse-driver as it will return directly a byte string.
    """

    IPv6NumToString = "IPv6NumToString"
    """
    Human string representation of the IPv6.
    With clickhouse-driver each address will be decoded as Python string.
    """


@dataclass(frozen=True)
class StoragePolicy:
    """
    See
    - https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-ttl
    - https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-multiple-volumes
    """

    name: str = "default"
    """Name of the ClickHouse storage policy to use for the table."""
    archive_to: str = "default"
    """Name of the ClickHouse archive volume."""
    archive_on: datetime = datetime(2100, 1, 1)
    """Date at which the table will be moved to the archive volume."""


@dataclass(frozen=True)
class Query:
    addr_type: AddrType = AddrType.IPv6

    @property
    def name(self) -> str:
        return self.__class__.__name__

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        # As a query user, prefer calling ``statements`` instead of ``statement`` as there
        # is no guarantees that the query will implement this method and return a single statement.
        raise NotImplementedError

    def statements(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> Sequence[str]:
        # Override this method if you want your query to return multiple statements.
        return (self.statement(measurement_id, subset),)

    def execute(
        self,
        url: str,
        measurement_id: str,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
    ) -> List:
        return [row for row in self.execute_iter(url, measurement_id, subsets, limit)]

    async def execute_async(
        self,
        url: str,
        measurement_id: str,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
    ) -> List:
        return [
            row
            async for row in self.execute_iter_async(
                url, measurement_id, subsets, limit
            )
        ]

    def execute_iter(
        self,
        url: str,
        measurement_id: str,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
    ) -> Iterator:
        with client(url) as c:
            for subset in subsets:
                for i, statement in enumerate(self.statements(measurement_id, subset)):
                    if limit:
                        statement += f"\nLIMIT {limit[0]} OFFSET {limit[1]}"
                    with LoggingTimer(
                        logger,
                        f"query={self.name}#{i} measurement_id={measurement_id} subset={subset}",
                    ):
                        rows = c.execute_iter(statement, settings=CH_QUERY_SETTINGS)
                        for row in rows:
                            yield row

    async def execute_iter_async(
        self,
        url: str,
        measurement_id: str,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
    ) -> AsyncIterator:
        async with async_client(url) as c:
            for subset in subsets:
                for i, statement in enumerate(self.statements(measurement_id, subset)):
                    if limit:
                        statement += f"\nLIMIT {limit[0]} OFFSET {limit[1]}"
                    with LoggingTimer(
                        logger,
                        f"query={self.name}#{i} measurement_id={measurement_id} subset={subset} limit={limit}",
                    ):
                        rows = await c.execute_iter(
                            statement, settings=CH_QUERY_SETTINGS
                        )
                        async for row in rows:
                            yield row

    def execute_http(
        self,
        url: str,
        measurement_id: str,
        database: str = "default",
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
    ) -> Iterator[dict]:
        for subset in subsets:
            for i, statement in enumerate(self.statements(measurement_id, subset)):
                if limit:
                    statement += f"\nLIMIT {limit[0]} OFFSET {limit[1]}"
                statement += "\nFORMAT JSONEachRow"
                with LoggingTimer(
                    logger,
                    f"query={self.name}#{i} measurement_id={measurement_id} database={database} subset={subset} limit={limit}",
                ):
                    r = httpx.get(
                        url, params={"query": statement, "database": database}
                    )
                    for line in r.content.splitlines():
                        yield json.loads(line)

    async def execute_http_async(
        self,
        url: str,
        measurement_id: str,
        database: str = "default",
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
    ) -> AsyncIterator[dict]:
        async with httpx.AsyncClient() as c:
            for subset in subsets:
                for i, statement in enumerate(self.statements(measurement_id, subset)):
                    if limit:
                        statement += f"\nLIMIT {limit[0]} OFFSET {limit[1]}"
                    statement += "\nFORMAT JSONEachRow"
                    with LoggingTimer(
                        logger,
                        f"query={self.name}#{i} measurement_id={measurement_id} database={database} subset={subset} limit={limit}",
                    ):
                        r = await c.get(
                            url, params={"query": statement, "database": database}
                        )
                        for line in r.content.splitlines():
                            yield json.loads(line)

    async def execute_concurrent(
        self,
        url: str,
        measurement_id: str,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
        concurrent_requests: int = (os.cpu_count() or 2) // 2,
    ) -> None:
        semaphore = Semaphore(concurrent_requests)

        async def do(subset: IPNetwork) -> None:
            async with semaphore:
                try:
                    await self.execute_async(url, measurement_id, (subset,), limit)
                except ServerException as e:
                    logger.error(
                        "query=%s subset=%s exception=%s", self.name, subset, e
                    )
                    raise e

        logger.info("query=%s concurrent_requests=%s", self.name, concurrent_requests)
        await asyncio.gather(*[do(subset) for subset in subsets])

    def _addr_cast(self, column: str) -> str:
        """Returns the column casted to the specified address type."""
        if self.addr_type == AddrType.IPv6:
            return column
        elif self.addr_type == AddrType.String:
            return f"CAST({column} AS String)"
        elif self.addr_type == AddrType.FixedString:
            return f"CAST({column} AS FixedString(16))"
        elif self.addr_type == AddrType.IPv6NumToString:
            return f"IPv6NumToString({column})"
        else:
            raise AttributeError("`addr_type` must be `AddrType` type")


@dataclass(frozen=True)
class FlowsQuery(Query):
    round_eq: Optional[int] = None
    "If specified, keep only the flows from this round."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the flows table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class LinksQuery(Query):
    filter_inter_round: bool = False
    "If true, exclude links inferred across rounds."

    filter_partial: bool = False
    "If true, exclude partial links: ``('::', node)`` and ``(node, '::')``."

    filter_virtual: bool = False
    "If true, exclude virtual links: ``('::', '::')``."

    near_or_far_addr: Optional[str] = None
    "If specified, keep only the links that contains this IP address."

    probe_protocol: Optional[int] = None
    "If specified, keep only the links inferred from probes sent with this protocol."

    probe_src_addr: Optional[str] = None
    """
    If specified, keep only the links inferred from probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    round_eq: Optional[int] = None
    "If specified, keep only the links from this round."

    round_leq: Optional[int] = None
    "If specified, keep only the links from this round or before."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the links table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        if self.round_eq:
            s += [eq("near_round", self.round_eq), eq("far_round", self.round_eq)]
        if self.round_leq:
            s += [leq("near_round", self.round_leq), leq("far_round", self.round_leq)]
        if self.near_or_far_addr:
            s += [
                or_(
                    ip_eq("near_addr", self.near_or_far_addr),
                    ip_eq("far_addr", self.near_or_far_addr),
                )
            ]
        if self.filter_inter_round:
            s += [not_("is_inter_round")]
        if self.filter_partial:
            s += [not_("is_partial")]
        if self.filter_virtual:
            s += [not_("is_virtual")]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class PrefixesQuery(Query):
    probe_protocol: Optional[int] = None
    "If specified, keep only the links inferred from probes sent with this protocol."

    probe_src_addr: Optional[str] = None
    """
    If specified, keep only the links inferred from probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the prefixes table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class ProbesQuery(Query):
    probe_protocol: Optional[int] = None
    "If specified, keep only probes sent with this protocol."

    round_eq: Optional[int] = None
    "If specified, keep only the probes from this round."

    round_geq: Optional[int] = None
    "If specified, keep only the probes from this round or after."

    round_leq: Optional[int] = None
    "If specified, keep only the probes from this round or before."

    round_lt: Optional[int] = None
    "If specified, keep only the probes from before this round."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the probes table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        if self.round_geq:
            s += [geq("round", self.round_geq)]
        if self.round_lt:
            s += [lt("round", self.round_lt)]
        if self.round_leq:
            s += [leq("round", self.round_leq)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class ResultsQuery(Query):
    filter_destination_host: bool = True
    "If true, ignore the replies from the destination host."

    filter_destination_prefix: bool = True
    "If true, ignore the replies from the destination prefix."

    filter_private: bool = True
    "If true, ignore the replies from private IP addresses."

    filter_invalid_probe_protocol: bool = True
    "If true, ignore the replies with probe protocol ≠ ICMP, ICMPv6 or UDP."

    time_exceeded_only: bool = True
    "If true, ignore non ICMP time exceeded replies."

    probe_protocol: Optional[int] = None
    "If specified, keep only the replies to probes sent with this protocol."

    probe_src_addr: Optional[str] = None
    """
    If specified, keep only the replies to probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    round_eq: Optional[int] = None
    "If specified, keep only the replies from this round."

    round_leq: Optional[int] = None
    "If specified, keep only the replies from this round or before."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the results table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        if self.round_leq:
            s += [leq("round", self.round_leq)]
        if self.filter_destination_host:
            s += [not_("destination_host_reply")]
        if self.filter_destination_prefix:
            s += [not_("destination_prefix_reply")]
        if self.filter_private:
            s += [not_("private_probe_dst_prefix"), not_("private_reply_src_addr")]
        if self.time_exceeded_only:
            s += ["time_exceeded_reply"]
        if self.filter_invalid_probe_protocol:
            s += ["valid_probe_protocol"]
        return reduce(and_, s or ["1"])



from diamond_miner.typing import FlowMapper

class SequentialFlowMapper(FlowMapper):
    def __init__(self, prefix_size: int = ...): ...

class IntervalFlowMapper(FlowMapper):
    def __init__(self, prefix_size: int = ..., step: int = ...): ...

class ReverseByteFlowMapper(FlowMapper): ...

class RandomFlowMapper(FlowMapper):
    def __init__(self, seed: int, prefix_size: int = ...): ...



from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import ProbesQuery, probes_table
from diamond_miner.typing import IPNetwork


class DeleteProbes(ProbesQuery):
    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        ALTER TABLE {probes_table(measurement_id)}
        DELETE WHERE {self.filters(subset)}
        SETTINGS mutations_sync = 1
        """



import asyncio
import os
import random
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Tuple

from zstandard import ZstdCompressor

from diamond_miner.defaults import DEFAULT_PROBE_DST_PORT, DEFAULT_PROBE_SRC_PORT
from diamond_miner.format import format_probe
from diamond_miner.logging import logger
from diamond_miner.queries import GetMDAProbes
from diamond_miner.rounds.mda import mda_probes
from diamond_miner.subsets import subsets_for
from diamond_miner.typing import FlowMapper, IPNetwork, Probe


async def mda_probes_parallel(
    filepath: Path,
    url: str,
    measurement_id: str,
    previous_round: int,
    mapper_v4: FlowMapper,
    mapper_v6: FlowMapper,
    probe_src_port: int = DEFAULT_PROBE_SRC_PORT,
    probe_dst_port: int = DEFAULT_PROBE_DST_PORT,
    adaptive_eps: bool = False,
    target_epsilon: float = 0.05,
    n_workers: int = (os.cpu_count() or 2) // 2,
) -> int:
    """
    Compute the probes to send given the previously discovered links.
    This function shuffle the probes on-disk using the following algorithm:
    https://lemire.me/blog/2010/03/15/external-memory-shuffling-in-linear-time/
    """
    loop = asyncio.get_running_loop()

    # NOTE: Make sure that the parameter of GetNextRound are equal to the
    # ones defines in mda_probes, in order to guarantee optimal subsets.
    subsets = await subsets_for(
        GetMDAProbes(
            adaptive_eps=adaptive_eps,
            round_leq=previous_round,
            filter_virtual=True,
            filter_inter_round=True,
            target_epsilon=target_epsilon,
        ),
        url,
        measurement_id,
    )

    if not subsets:
        return 0

    n_files_per_subset = 8192 // len(subsets)

    logger.info(
        "mda_probes n_workers=%s n_subsets=%s n_files_per_subset=%s",
        n_workers,
        len(subsets),
        n_files_per_subset,
    )

    with TemporaryDirectory(dir=filepath.parent) as temp_dir:
        with ProcessPoolExecutor(n_workers) as pool:
            futures = [
                loop.run_in_executor(
                    pool,
                    worker,
                    Path(temp_dir) / f"subset_{i}",
                    url,
                    measurement_id,
                    previous_round,
                    mapper_v4,
                    mapper_v6,
                    probe_src_port,
                    probe_dst_port,
                    adaptive_eps,
                    target_epsilon,
                    subset,
                    n_files_per_subset,
                )
                for i, subset in enumerate(subsets)
            ]
            n_probes = sum(await asyncio.gather(*futures))

        logger.info("mda_probes status=merging")
        file_list = Path(temp_dir) / "files.txt"
        proc = await asyncio.create_subprocess_shell(
            f"find {temp_dir} -name 'subset_*.csv.zst' | shuf > {file_list}"
        )
        await proc.wait()
        proc = await asyncio.create_subprocess_shell(
            f"xargs cat < {file_list} > {filepath}"
        )
        await proc.wait()

    return n_probes


def worker(
    prefix: Path,
    url: str,
    measurement_id: str,
    round_: int,
    mapper_v4: FlowMapper,
    mapper_v6: FlowMapper,
    probe_src_port: int,
    probe_dst_port: int,
    adaptive_eps: bool,
    target_epsilon: float,
    subset: IPNetwork,
    n_files: int,
) -> int:
    """
    Execute the :class:`diamond_miner.queries.GetNextRound` query
    on the specified subset, and write the probes to the specified file.
    """
    # TODO: random.shuffle is slow...
    # A potentially simpler and better way would be to shuffle
    # the result of the next round query, which is small (~10M rows),
    # and to directly write the probes to a file depending on hash(flow_id).
    # e.g. ORDER BY rand() in GetNextRound.

    # The larger `max_probes_in_memory`, the better the performance
    # (less calls to `probes.clear()`)
    # and the randomization but the more the memory usage.
    max_probes_in_memory = 1_000_000

    outputs: List[Tuple] = []
    for i in range(n_files):
        ctx = ZstdCompressor(level=1)
        file = prefix.with_suffix(f".{i}.csv.zst").open("wb")
        stream = ctx.stream_writer(file)
        outputs.append((ctx, file, stream))

    probes_by_file: List[List[Probe]] = [[] for _ in range(n_files)]
    n_probes = 0

    for probe in mda_probes(
        url,
        measurement_id,
        round_,
        mapper_v4,
        mapper_v6,
        probe_src_port,
        probe_dst_port,
        adaptive_eps,
        target_epsilon,
        (subset,),
    ):
        # probe[:-2] => (probe_dst_addr, probe_src_port, probe_dst_port)
        probes_by_file[hash(probe[:-2]) % n_files].append(probe)
        n_probes += 1
        if n_probes % max_probes_in_memory == 0:
            flush(probes_by_file, outputs)

    flush(probes_by_file, outputs)

    for ctx, file, stream in outputs:
        stream.close()
        file.close()

    return n_probes


def flush(probes_by_file: List[List[Probe]], outputs: List[Tuple]) -> None:
    for file_id, probes in enumerate(probes_by_file):
        _, _, stream = outputs[file_id]
        random.shuffle(probes)
        for probe in probes:
            stream.write(format_probe(*probe).encode("ascii") + b"\n")
        probes.clear()



from math import ceil, log


def stopping_point(k: int, eps: float = 0.05) -> int:
    """
    Return the number `n_k` of probes that guarantees that the probability of not
    detecting `k` outgoing load-balanced edges is lower than `eps`.
    Recurrent formula in [1], improved direct formula in [2].
    [1] "Failure Control in Multipath Route Tracing"
    https://www.researchgate.net/publication/224500381_Failure_Control_in_Multipath_Route_Tracing
    [2] "Collecter un nombre inconnu de coupons"
    https://hal.inria.fr/hal-01787252/document
    >>> stopping_point(1, 0.05)
    0
    >>> stopping_point(2, 0.05)
    6
    >>> stopping_point(3, 0.05)
    11
    >>> stopping_point(11, 0.05)
    57
    >>> stopping_point(101, 0.05)
    765

    NOTE: There is a typo in the D-Miner paper: n(101) = 765, not 757.
    """
    assert (k >= 1) and (0 <= eps <= 1)
    if k == 1:
        return 0
    return ceil(log(eps / k) / log((k - 1) / k))



from typing import Any, Iterator, Optional, Sequence, Tuple, Union

class ParameterGrid:
    def __init__(self, *parameters: Sequence): ...
    def __getitem__(self, index: Union[Sequence[int], int]) -> Sequence[Any]: ...
    def __iter__(self) -> Iterator[Sequence[Any]]: ...
    def __len__(self) -> int: ...
    @property
    def size(self) -> int: ...
    def shuffled(
        self, rounds: int = ..., seed: Optional[int] = ...
    ) -> Iterator[Sequence[Any]]: ...
